<?php /** @noinspection DuplicatedCode */

use function LitePress\LP_Translate\can;

/**
 * Ajax "sync" route.
 * Extracts strings from source (POT or code) and returns to the browser for in-editor merge.
 */
class Loco_ajax_SyncController extends Loco_mvc_AjaxController {

    /**
     * {@inheritdoc}
     */
    public function render(){
        
        $post = $this->validate();
        
        $bundle = Loco_package_Bundle::fromId( $post->bundle );
        $project = $bundle->getProjectById( $post->domain );
        if( ! $project instanceof Loco_package_Project ){
            throw new Loco_error_Exception('No such project '.$post->domain);
        }

        // Merging on back end is only required if existing target file exists.
        // Currently it always will and the editor is not permitted to contain unsaved changes when syncing.
        if( ! $post->has('path') ){
            throw new Loco_error_Exception('path argument required');
        }

        // 确认当前用户是否对该项目拥有管理权限(manage)
        $can_manage = can( 'manage', $post->domain );

        /**
         * 如果当前用户对当前项目拥有管理权限，则不对翻译模板同步改为对当前翻译同步
         *
         * 这样做是因为，对于管理员来讲，翻译平台上的项目原文需要他自己上传，如果他还要从平台同步模板就死循环了
         * 所以他必须从本地提取原文并同步到平台。同时平台上现存的翻译需要他同步到本地，所以这里将原文同步以获取翻译
         * 而原文会稍后与翻译模板合并，以实现从本地提取原文并同步翻译的目的
         */
        if ( $can_manage ) {
            $file = apply_filters( 'lp_translate_sync_file', false, $post->domain, $post->path );
        }

        if ( empty( $file ) ) {
            $file = new Loco_fs_File( $post->path );
            $base = loco_constant('WP_CONTENT_DIR');
            $file->normalize($base);
        }
        $target = Loco_gettext_Data::load($file);

        // POT file always synced with source code
        $type = $post->type;
        if( 'pot' === $type ){
            $potfile = null;
        }
        // allow front end to configure source file. (will have come from $target headers)
        else if( $post->sync ){
            $potfile = new Loco_fs_File( $post->sync );
            $potfile->normalize($base);
        }
        // else use project-configured template path (must return a file)
        else {
            /**
             * 该钩子用来以不同方式来获取翻译模板
             */
            if ( ! $can_manage ) {
                $potfile = apply_filters( 'lp_translate_sync_file', false, $post->domain, $post->path );
            }

            if ( empty( $potfile ) ) {
                $potfile = $project->getPot();
            }
        }
        
        // keep existing behaviour when template is missing, but add warning according to settings.
        if( $potfile && ! $potfile->exists() ){
            $conf = Loco_data_Settings::get()->pot_expected;
            if( 2 === $conf ){
                throw new Loco_error_Exception('Plugin settings disallow missing templates');
            }
            if( 1 === $conf ){
                // 不再弹出该提示
                // Translators: %s will be replaced with the name of a missing POT file
                //Loco_error_AdminNotices::warn( sprintf( __('Falling back to source extraction because %s is missing','lp-translate'), $potfile->basename() ) );
            }
            $potfile = null;
        }
        
        // defaults: no msgstr and no json
        $translate = false;
        $syncjsons = array();
        
        // Parse existing POT for source
        if( $potfile ){
            $this->set('pot', $potfile->basename() );
            try {
                $source = Loco_gettext_Data::load($potfile);
            }
            catch( Exception $e ){
                // translators: Where %s is the name of the invalid POT file
                throw new Loco_error_ParseException( sprintf( __('Translation template is invalid (%s)','lp-translate'), $potfile->basename() ) );
            }
            // Only copy msgstr fields from source if it's a user-defined PO template and "copy translations" was selected.
            $opts = new Loco_gettext_SyncOptions( new LocoPoHeaders );
            $opts->setSyncMode( $post->mode );
            if( 'pot' !== $potfile->extension() ){
                $translate = $opts->mergeMsgstr();
            }
            // related JSONs will only wor if source is a localized PO.
            if( $opts->mergeJson() ){
                $siblings = new Loco_fs_Siblings($potfile);
                $syncjsons = $siblings->getJsons();
            }
        }
        // else extract POT from source code
        else {
            $this->set('pot', '' );
            $domain = (string) $project->getDomain();
            $extr = new Loco_gettext_Extraction($bundle);
            $extr->addProject($project);
            // bail if any files were skipped
            if( $list = $extr->getSkipped() ){
                $n = count($list);
                $maximum = Loco_mvc_FileParams::renderBytes( wp_convert_hr_to_bytes( Loco_data_Settings::get()->max_php_size ) );
                $largest = Loco_mvc_FileParams::renderBytes( $extr->getMaxPhpSize() );
                // Translators: (1) Number of files (2) Maximum size of file that will be included (3) Size of the largest encountered
                $text = _n('%1$s file has been skipped because it\'s %3$s. (Max is %2$s). Check all strings are present before saving.','%1$s files over %2$s have been skipped. (Largest is %3$s). Check all strings are present before saving.',$n,'lp-translate');
                $text = sprintf( $text, number_format($n), $maximum, $largest );
                // not failing, just warning. Nothing will be saved until user saves editor state
                Loco_error_AdminNotices::warn( $text );
            }
            // Have source strings. These cannot contain any translations.
            $source = $extr->includeMeta()->getTemplate($domain);
        }
        
        // establish on back end what strings will be added, removed, and which could be fuzzy-matches
        $matcher = new Loco_gettext_Matcher;
        $matcher->loadRefs($source,$translate);
        // merging JSONs must be done before fuzzy matching as it may add source strings
        $matcher->loadJsons($syncjsons);
        // Fuzzy matching only applies to syncing PO files. POT files will always do hard sync (add/remove)
        if( 'po' === $type ){
            $fuzziness = Loco_data_Settings::get()->fuzziness;
            $matcher->setFuzziness( (string) $fuzziness );
        }
        else {
            $matcher->setFuzziness('0');
        }
        // update matches sources, deferring unmatched for deferred fuzzy match 
        $merged = clone $target;
        $merged->clear();
        $this->set( 'done', $matcher->merge($target,$merged) );
        $merged->sort();
        $this->set( 'po', $merged->jsonSerialize() );
        
        return parent::render();
    }
    
    
}